%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% DFBAlab: Dynamic Flux Balance Analysis laboratory                       %
% Process Systems Engineering Laboratory, Cambridge, MA, USA              %
% July 2014                                                               %
% Written by Jose A. Gomez and Kai Hffner                                %
%                                                                         % 
% This code can only be used for academic purposes. When using this code  %
% please cite:                                                            %
%                                                                         %
% Gomez, J.A., Hffner, K. and Barton, P. I.                              %
% DFBAlab: A fast and reliable MATLAB code for Dynamic Flux Balance       %
% Analysis. Submitted.                                                    %
%                                                                         %
% COPYRIGHT (C) 2014 MASSACHUSETTS INSTITUTE OF TECHNOLOGY                %
%                                                                         %
% Read the LICENSE.txt file for more details.                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Clear workspace

clear all

%% Paths to LP solver and DFBAlab
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
addpath(genpath('C:\Gurobi605'))
addpath(genpath('C:\Users\Mike\Documents\MATLAB\DFBAlab_0'))
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Problem parameters
%
% INFO: You can use the INFO struct to pass parameters. Don't use any of 
% the names already declared or: INFO.t (carries time information), 
% INFO.ncost, INFO.lexID, INFO.LlexID, INFO.lbct, INFO.ubct, INFO.sense,
% INFO.b, INFO.pair. 
%
% nspecies = number of species
% ns = number of extracellular states
% npoints = number of discretization points
% tspan = simulation time
% icond = 1 if loading initial condition from file
%       = 0 if specifying initial condition from bulk concentrations
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
nspecies = 3;
ns = 21;
npoints = 25;
tspan = [0,10000];
icond = 0;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Calculate and store problem parameters

N = npoints-1;
nmet = ns-nspecies-1;
INFO.npoints = npoints;
INFO.N = N;
INFO.ns = ns;
INFO.nspecies = nspecies;
INFO.nmet = nmet;

%% Load models
%
% These should be .mat files generated by the COBRA toolbox. 
% When generating these files using the COBRA toolbox, a big number is used
% as infinity. This number should be fed to the DB vector (Default bound).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Species 1 = CD; Species 2 = BT, Species 3 = FP, Species 4 = EC
%
load BT_model
BT_biofilm.lb(269) = 8.43;
species{1} = BT_biofilm;
load FP_model
FP_biofilm.lb(163) = 8.43;
species{2} = FP_biofilm;
load EC_model
EC_biofilm.lb(375) = 6.75;
species{3} = EC_biofilm;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Specify and store species models and upper bounds

nmodel = nspecies*npoints;
INFO.nmodel = nmodel;
for j = 1:nspecies
    for i = 1:npoints
        model{(j-1)*npoints+i} = species{j};
        DB((j-1)*npoints+i) = 1000;
    end
end
INFO.DB = DB;

%% exID array
%
% You can either search the reaction names by name or provide them directly
% in the exID array.
% RxnNames = {'EX_glc(e)', 'EX_ac(e)', 'biomass'};
% for i = 1:length(RxnNames)
%    [a,exID(i)] = ismember(RxnNames(i),model.rxns);
% end
%
% Lower bounds and upper bounds for these reactions should be provided in
% the RHS code.
%
% This codes solves the LPs in standard form. Bounds on exchange fluxes in 
% the exID array can be modified directly on the first 2*n rows where n is 
% the number of exchange fluxes. Order will be lower bound, upper bound, 
% lower bound, upper bound in the same order as exID. 
%
% NOTE: All bounds on fluxes in the exID arrays are relaxed to -Inf and 
% + Inf. These bounds need to be updated if needed in the RHS file.
%
% neflux = number of exchange fluxes for each species
% eflux = ordered exchange fluxes for each species as separate row
% eflag = 1 if uptake fluxes are negative for the species (normal)
%       = 0 if uptake fluxes are positive for the species (unusual)
% edir = -1 if eflag = 0 and eflux element is an uptake flux 
%       = 1 otherwise   
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% eflux: Biomass, glucose, cysteine, isoleucine, leucine, methionine, 
%        proline, serine, tryptophan, valine, acetate, butyrate, co2,
%        ethanol, formate, lactate, propionate, succinate
% BT: "butyrate" (replace with lactose)
% FP: "ethanol", "propionate" (replace with lactose and methanol)
% 
neflux = [18 18 18 18];
eflux = [281 559 532 607 627 652 684 702 718 724 471 626 512 542 548 619 683 710
         175 359 346 385 395 404 424 433 443 448 306 332 342 394 353 393 402 437
         1005 849 786 885 896 913 950 961 981 998 729 763 774 810 824 893 946 967];
eflag = [1 1 1]';
edir = [ones(1,neflux(1));
        ones(1,neflux(2))
        ones(1,neflux(3))];
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Specify and store exchange fluxes

for j = 1:nspecies
    for i = 1:npoints
        exID{(j-1)*npoints+i} = eflux(j,:);
    end
end
INFO.neflux = neflux;
INFO.eflux = eflux;
INFO.eflag = eflag;
INFO.edir = edir;
INFO.exID = exID;

%% Cost vectors
%
% Usually the first cost vector will be biomass maximization, but it can
% be any other objective. The CPLEX objects will minimize by default. 
% Report only nonzero elements. 
% The structure should be:
% C{model} = struct
% Element C{k}(i) of C, is the cost structure i for model k. 
% C{k}(i).sense = +1 for minimize, or -1 for maximize.
% C{k}(i).rxns = array containing the reactions in this objective. 
% C{k}(i).wts = array containing coefficients for reactions reported in 
% rxns. Both arrays should have the same length. 
% Example, if:
% C{k}(i).rxns = [144, 832, 931];
% C{k}(i).wts = [3, 1, -1];
% Then the cost vector for this LP will be:
% Cost{k}(i) = 3*v_144 + v_832 - v_931 (fluxes for model k). 
% This cost vector will be either maximized or minimized depending on the
% value of C{k}(i).sense.
%
% In SBML files, usually production fluxes are positive and uptake fluxes
% are negative. Keep in mind that maximizing a negative flux implies 
% minimizing its absolute value.
% Different models can have different number of objectives. 
%
% cflux = ordered exchange fluxes for costing with each species as separate row
% cobj = ordered optimization objectives for costing with each species as separate row
% wobj = ordered optimization weights for costing with each species as separate row
% cdir = -1 if eflag = 0 and cflux element is an uptake flux 
%       = 1 otherwise  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% BT cflux: biomass, acetate, succinate, propionate, co2, lactate, ethanol, formate, 
%           glucose, cysteine, isoleucine, leucine, methionine, proline,
%           serine, tryptophan, valine, "butyrate"
% FP cflux: biomass, lactate, butyrate, formate, co2, acetate, succinate,
%           glucose, cysteine, isoleucine, leucine, methionine, proline,
%           serine, tryptophan, valine, "ethanol", "propionate"
% EC cflux: biomass, acetate, ethanol, formate, lactate, succinate, co2, butyrate, 
%           propionate, glucose, cysteine, isoleucine, leucine, 
%           methionine, proline, serine, tryptophan, valine
%
cflux = [281 471 710 683 512 619 542 548 559 532 607 627 652 684 702 718 724 626
        175 393 332 353 342 306 437 359 346 385 395 404 424 433 443 448 394 402
        1005 729 810 824 893 967 774 763 946 849 786 885 896 913 950 961 981 998];
minim = 1;
maxim = -1;
cobj = [maxim maxim maxim maxim maxim maxim maxim maxim maxim maxim maxim maxim maxim maxim maxim maxim maxim minim
        maxim maxim maxim maxim maxim maxim maxim maxim maxim maxim maxim maxim maxim maxim maxim maxim minim minim
        maxim maxim maxim maxim maxim maxim maxim maxim maxim maxim maxim maxim maxim maxim maxim maxim maxim maxim]; 
 wobj = [ones(1,neflux(1))
        ones(1,neflux(2))
        ones(1,neflux(3))];
cdir = [ones(1,neflux(1))
        ones(1,neflux(2))
        ones(1,neflux(3))];
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
% Store cost information

 for j=1:nspecies
    for i=1:npoints
        for k=1:neflux(j)
            C{(j-1)*npoints+i}(k).sense = cobj(j,k);
            C{(j-1)*npoints+i}(k).rxns = cflux(j,k);
            C{(j-1)*npoints+i}(k).wts = wobj(j,k);  
        end
    end
 end
 
 INFO.C = C;
 INFO.cflux = cflux;
 INFO.cdir = cdir;

%% Biofilm parameters
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Species: BT, FP, EC
% Metabolites: glucose, cysteine, isoleucine, leucine, methionine, proline,
%              serine, tryptophan, valine, acetate, butyrate, co2, ethanol,
%              formate, lactate, propionate, succinate
%

L = 40;                                        % Biofilm thickness, micron
x0 = [10 10 10];                               % initial biomass concentrations if icond=0, g/L
xdiff = [1e-10*ones(1,nspecies)];              % Biomass diffusion coefficients, cm^2/s
xmt = [1e-7*ones(1,nspecies)];                 % Biomass mass transfer coefficients, cm/s
kxd = [0 0 0];                                 % Basal cell death rate constants, 1/h
kxa = [1000 1000 1000];                        % Antibiotic mediated cell death rate constant, ?   
metb = [5 0 0 0 0.5 0 0.5 0.5 0 0*ones(1,8)];  % Bulk metabolite concentrations, mmol/L 
metdiff = 2e-6*ones(1,nmet);           	       % Metabolite diffusion coefficients, cm^2/s
metmt = [2e-4*ones(1,9) 0.5e-5*ones(1,8)];     % Metabolite mass transfer coefficients, cm/s
antb = 0*7.5e-3; %1.5e-3;                      % Bulk antibiotic concentration, mmol/L
antdiff = 1.5e-6; %6e-7;                       % Antibiotic diffusion coefficient, cm^2/s
antmt = 2e-4; %2e-4                        	   % Antibiotic mass transfer coefficient, cm/s
kax = 5; %0.5;                          	   % Antibiotic-cell binding constant, L/mmol/h
antdose = 12; %6;                         	   % Antibiotic dosage period, h
anttau = 6; %8;                           	   % Antibiotic time constant, h
antdur = 336; %336                        	   % Antibiotic treatment duration, h
antstr = 100;                         	       % Antibiotic starting time, h
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Store biofilm parameters

zs = L/N;
condit = [L,xdiff,xmt,kxd,kxa,metb,metdiff,metmt,antb,antdiff,antmt,kax,...
          antdose,anttau,antdur,antstr,zs];
INFO.condit=condit;

%% Nutrient uptake kinetics
%
% vmax = maximum uptake rate for metabolites with each species as a row
% km = uptake rate saturation constant for metabolites with each species as a row
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Metabolites: glucose, cysteine, isoleucine, leucine, methionine, proline,
%              serine, tryptophan, valine, acetate, butyrate, co2, ethanol,
%              formate, lactate, propionate, succinate
%
vmax = [% No crossfeeding
        %10 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0
        %10 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0
        %10 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0];
        % Maximal crossfeeding
        %10 1 1 1 1 1 1 1 1 10 0 0 10 10 10 10 10
        %10 1 1 1 1 1 1 1 1 10 10 10 0 10 10 0 10
        %10 1 1 1 1 1 1 1 1 10 10 10 10 10 10 10 10];
        % Miminal crossfeeding
        10 1 1 1 1 1 1 1 1 0 0 0 10 0 0 0 0 % 0.75, 0.80, 0.85, 0.90, base
        10 1 1 1 1 1 1 1 1 10 0 0 0 0 0 0 10 % Ace 6.25 6.50 6.75 7.00 base; Suc 3.25, 3.50, 3.75, 4.00, base
        10 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0];
km = [0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5
      0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5
      0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5];
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Store uptake kinetics

INFO.vmax = vmax;
INFO.km = km;

%% Initial conditions

if icond == 1

% Load initial conditions
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    load 3species_L40_NP25_BASE_IC;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    yo = Y(end,1:end-nmodel);
        
else
    
% Specify initial conditions

    yo = [];
    for i = 1:N+1
        yz = [x0,metb,antb];
        yo = [yo yz];
    end
    
end

% Add initial condition for penalty variable

yo = [yo zeros(1,nmodel)];
Y0 = yo;

%% CPLEX Objects construction parameters

INFO.LPsolver = 1; % CPLEX = 0, Gurobi = 1.
                   % CPLEX works equally fine with both methods.
                   % Gurobi seems to work better with Method = 1, and 
                   % Mosek with Method = 0.
INFO.tol = 1E-9; % Feasibility, optimality and convergence tolerance for Cplex (tol>=1E-9). 
                 % It is recommended it is at least 2 orders of magnitude
                 % tighter than the integrator tolerance. 
                 % If problems with infeasibility messages, tighten this
                 % tolerance.
INFO.tolPh1 = INFO.tol; % Tolerance to determine if a solution to phaseI equals zero.
                        % It is recommended to be the same as INFO.tol. 
INFO.tolevt = 2E-6; % Tolerance for event detection. Has to be greater
                          % than INFO.tol.

%% Integration options
%
% You can modify the integration tolerances here.
% If some of the flows become negative after running the simulation once
% you can add the 'Nonnegative' option.
% Include function to compute non-zero Jacobian entries.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
NN = 1:length(Y0);      % Do not modify
options = odeset('AbsTol',5E-7,'RelTol',5E-7,'NonNegative',NN,...
    'Events',@evts,'JPattern',jac_3species_biofilm(N,ns));
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Model setup

[model,INFO] = ModelSetupM(model,Y0,INFO);
INFO.INFtimes = zeros(1,nmodel);

if INFO.LPsolver == 0
    [INFO] = LexicographicOpt(model,INFO);
elseif INFO.LPsolver == 1
    [INFO] = LexicographicOptG(model,INFO);
else
    display('Solver not currently supported.');
end

%% Integration loop

tint = 0;
TF = [];
YF = [];
xflux = [];
mflux = [];
while tint<tspan(2)
    INFO.INFtimes
    
% Look at MATLAB documentation if you want to change solver.
% ode15s is more or less accurate for stiff problems.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    [T,Y] = ode15s(@odes_3species_biofilm,tspan,Y0,options,INFO);
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    TF = [TF;T];
    YF = [YF;Y];
    for i=1:length(T)
        [dy,xflux1,mflux1] = odes_3species_biofilm(T(i),Y(i,:),INFO);
        xflux = [xflux;xflux1];
        mflux = [mflux;mflux1];
    end
    tint = T(end);
    tspan = [tint,tspan(2)];
    Y0 = Y(end,:);
    if tint == tspan(2)
        break;
    end
    
% Update b vector 
    
    [INFO] = bupdate(tint,Y0,INFO);

% Determine model with basis change
    
    value = evts(tint,Y0,INFO);
    ind = find(value(1:end-nmodel)<=0);
    fprintf('Basis change at time %d. ',tint);
    k = 0;
    ct = 0;
    
% Detect model infesibilities

    INFvector =  value(end-nmodel+1:end);
    for i=1:length(INFvector)
        if INFvector(i)>0 && INFO.INFtimes(i) == 0
            INFO.INFtimes(i) = tint;
            fprintf('Model %i is now dying at time %d. \n',i,tint);
        end
    end
    
    while ~isempty(ind)
        k = k + 1;
        ct = ct + size(model{k}.A,1);
        ind2 = find(ind<=ct);
        if ~isempty(ind2)
           INFO.flagbasis = k; 
           fprintf('Model %i. \n',k);
           
% Perform lexicographic optimization

           if INFO.LPsolver == 0
               [INFO] = LexicographicOpt(model,INFO);
           elseif INFO.LPsolver == 1
               [INFO] = LexicographicOptG(model,INFO);
           else
               display('Solver not currently supported.');
           end
           ind(ind2)=[];
        end
    end
end

T = TF;
Y = YF;
       
%% Plot results
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
pflag = plot_3species_biofilm(T,Y,xflux,mflux,nspecies,nmet,N,ns,zs);
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%