%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% DFBAlab: Dynamic Flux Balance Analysis laboratory                       %
% Process Systems Engineering Laboratory, Cambridge, MA, USA              %
% July 2014                                                               %
% Written by Jose A. Gomez and Kai Hffner                                %
%                                                                         % 
% This code can only be used for academic purposes. When using this code  %
% please cite:                                                            %
%                                                                         %
% Gomez, J.A., Hffner, K. and Barton, P. I.                              %
% DFBAlab: A fast and reliable MATLAB code for Dynamic Flux Balance       %
% Analysis. Submitted.                                                    %
%                                                                         %
% COPYRIGHT (C) 2014 MASSACHUSETTS INSTITUTE OF TECHNOLOGY                %
%                                                                         %
% Read the LICENSE.txt file for more details.                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [dy,xflux1,mflux1] = odes_3species_biofilm(t, y, INFO)

t

% Set problem parameters

nmodel = INFO.nmodel;
N = INFO.N;
ns = INFO.ns;
nspecies = INFO.nspecies;
neflux = INFO.neflux;
nmet = INFO.nmet;
npoints = INFO.npoints;
condit = INFO.condit;
edir = INFO.edir;
eflux = INFO.eflux;
cflux = INFO.cflux;

% Set biofilm parameters

L = condit(1);
bmax = condit(2);
xdiff = 3600*(1e6)^2/(100)^2*condit(3:3+nspecies-1)';
kxd = condit(3+nspecies:3+2*nspecies-1)';
kxa = condit(3+2*nspecies:3+3*nspecies-1)';
metb = condit(3+3*nspecies:3+3*nspecies+nmet-1)';
metdiff = 3600*(1e6)^2/(100)^2*condit(3+3*nspecies+nmet:3+3*nspecies+2*nmet-1)';
metmt = 3600*1e6/100*condit(3+3*nspecies+2*nmet:3+3*nspecies+3*nmet-1)';
antb = condit(3+3*nspecies+3*nmet);
antdiff = 3600*(1e6)^2/(100)^2*condit(4+3*nspecies+3*nmet);   
antmt = 3600*1e6/100*condit(5+3*nspecies+3*nmet);             
kax = condit(6+3*nspecies+3*nmet);
antdose = condit(7+3*nspecies+3*nmet);
anttau = condit(8+3*nspecies+3*nmet);
antdur = condit(9+3*nspecies+3*nmet);
antstr = condit(10+3*nspecies+3*nmet);
zs = condit(11+3*nspecies+3*nmet);

L = condit(1);
xdiff = 3600*(1e6)^2/(100)^2*condit(2:2+nspecies-1)';
xmt = 3600*1e6/100*condit(2+nspecies:2+2*nspecies-1)';
kxd = condit(2+2*nspecies:2+3*nspecies-1)';
kxa = condit(2+3*nspecies:2+4*nspecies-1)';
metb = condit(2+4*nspecies:2+4*nspecies+nmet-1)';
metdiff = 3600*(1e6)^2/(100)^2*condit(2+4*nspecies+nmet:2+4*nspecies+2*nmet-1)';
metmt = 3600*1e6/100*condit(2+4*nspecies+2*nmet:2+4*nspecies+3*nmet-1)';
antb = condit(2+4*nspecies+3*nmet);
antdiff = 3600*(1e6)^2/(100)^2*condit(3+4*nspecies+3*nmet);   
antmt = 3600*1e6/100*condit(4+4*nspecies+3*nmet);             
kax = condit(5+4*nspecies+3*nmet);
antdose = condit(6+4*nspecies+3*nmet);
anttau = condit(7+4*nspecies+3*nmet);
antdur = condit(8+4*nspecies+3*nmet);
antstr = condit(9+4*nspecies+3*nmet);
zs = condit(10+4*nspecies+3*nmet);

% Set extracellular state variables

for i=1:N+1
    xs(:,i) = y((i-1)*ns+1:(i-1)*ns+nspecies);
    ms(:,i) = y((i-1)*ns+nspecies+1:(i-1)*ns+nspecies+nmet);
    as(i) = y((i-1)*ns+nspecies+nmet+1);
end

% The elements of the flux matrix have the sign given to them by the
% coefficients in the Cost vector in main. 
% Example, if:
% C{k}(i).rxns = [144, 832, 931];
% C{k}(i).wts = [3, 1, -1];
% Then the cost vector for this LP will be:
% flux(k,i) = 3*v_144 + v_832 - v_931 
% The penalty is an array containing the feasibility LP problem optimal
% objective function value for each model. 

% Update bounds and solve for fluxes

[flux,penalty] = solveModel(t,y,INFO);
pflux = flux;
xflux1 = [];
mflux1 = [];

% Check for LP infeasbility

for i=1:N+1
    
    if y(end-nmodel+i) > 1e-6
        penalty(i) = 0;
    end
    
    if y(end-nmodel+N+1+i) > 1e-6
        penalty(N+1+i) = 0;
    end
    
    if y(end-nmodel+2*(N+1)+i) > 1e-6
        penalty(2*(N+1)+i) = 0;
    end
    
end

% Set exchange fluxes
     
for i = 1:nspecies
    for j = 1:neflux(i)
        idx(i,j) =  find(ismember(cflux(i,:),eflux(i,j)));
    end
end 

for i = 1:nspecies
    xe(i,:) =  pflux((i-1)*npoints+1:i*npoints,idx(i,1));
end

for i = 1:nspecies
    for j = 1:nmet
        me(i,j,:) =  pflux((i-1)*npoints+1:i*npoints,idx(i,j+1));
    end
end

%me(1,10,:) = 0;   % Eliminate BT acetate synthesis
%me(1,17,:) = 0;   % Eliminate BT succinate synthesis
%me(3,10,:) = 0;   % Eliminate EC acetate synthesis
%me(3,17,:) = 0;   % Eliminate EC succinate synthesis

xflux1 = [xflux1 xe];
mflux1 = [mflux1 me];

% Calculate antibiotic bulk concentration

if t >= antstr
    if t < antdur+antstr
        tb = mod(t,antdose);
        antb = antb*exp(-tb/anttau);
        %metb(1) = metb(1)*exp(-(t-100)/anttau);
    else
        antb = 0;
        %metb(1) = metb(1)*(1-exp(-(t-436)/anttau));
    end
else
    antb = 0;
    %metb(1) = 10;
end

% Calculate lower boundary conditions

x0 = xs(nspecies+1:2*nspecies)'+2.*xmt.*zs.*(0-xs(1:nspecies)')./xdiff;
m0 = ms(nmet+1:2*nmet)'+2.*metmt.*zs.*(metb-ms(1:nmet)')./metdiff;
a0 = as(2)+2*antmt*zs*(antb-as(1))/antdiff;

% Calculate upper boundary conditions

xf = xs((N-1)*nspecies+1:N*nspecies)';
mf = ms((N-1)*nmet+1:N*nmet)'+2.*metmt.*zs.*(0-ms(N*nmet+1:(N+1)*nmet)')./metdiff;
af = as(N);

% Evaluate equations

dy = [];

for i=1:N+1
    
% Compute second order derivatives

    if i==1
        xd2(:,1) = (xs(nspecies+1:2*nspecies)'-2.*xs(1:nspecies)'+x0)./zs^2;
        md2(:,1) = (ms(nmet+1:2*nmet)'-2.*ms(1:nmet)'+m0)./zs^2;
        ad2 = (as(2)-2*as(1)+a0)/zs^2;
              
    elseif i==N+1
        xd2(:,1) = (xf-2.*xs(N*nspecies+1:(N+1)*nspecies)'+xs((N-1)*nspecies+1:N*nspecies)')./zs^2;
        md2(:,1) = (mf-2.*ms(N*nmet+1:(N+1)*nmet)'+ms((N-1)*nmet+1:N*nmet)')./zs^2;
        ad2 = (af-2*as(N+1)+as(N))/zs^2;
        
    else
        xd2(:,1) = (xs(i*nspecies+1:(i+1)*nspecies)'-2.*xs((i-1)*nspecies+1:i*nspecies)'+xs((i-2)*nspecies+1:(i-1)*nspecies)')./zs^2;
        md2(:,1) = (ms(i*nmet+1:(i+1)*nmet)'-2.*ms((i-1)*nmet+1:i*nmet)'+ms((i-2)*nmet+1:(i-1)*nmet)')./zs^2;
        ad2 = (as(i+1)-2*as(i)+as(i-1))/zs^2;
        
    end
    
    % Evaluate RHS of ODEs
    
    xt = sum(xs((i-1)*nspecies+1:i*nspecies));
    dx =  xe(:,i).*xs(:,i)-kxd.*xs(:,i)-kxa.*xe(:,i).*xs(:,i).*as(i)+xdiff.*xd2(:,1);
    %dx = 0.*xe(:,i).*xs(:,i);
    dmet = metdiff.*md2(:,1);
    for j = 1:nspecies
        dmet = dmet+edir(j,2:end)'.*me(j,:,i)'.*xs(j,i);
    end
    dant = -kax*as(i)*xt+antdiff*ad2;
    dy = [dy dx' dmet' dant];

end

% Add penalty variable

dylast=zeros(1,nmodel);
for i=1:nmodel
    dylast(i) = penalty(i);
end
dy = [dy dylast];
dy = dy';
end

